/*
 * Copyright (c) 2017, Yuichiro MORIGUCHI
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * * Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * * Neither the name of the Yuichiro MORIGUCHI nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Yuichiro MORIGUCHI BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
jsmin = {};
(function(J) {
	var unaryop, stmtbase, exprbase;
	unaryop = function(priority, expr, word) {
		var r, spc = word ? ' ' : '';
		if(priority > expr.priority()) {
			r = '(' + expr.action() + ')';
		} else {
			r = spc + expr.action();
		}
		return r;
	};
	exprbase = function() {
		return {
			type: 'op',
			isCall: function() {
				return false;
			},
			isParen: function() {
				return false;
			},
			isFunc: function() {
				return false;
			}
		};
	};
	J.infixOperator = function(op, priority, assoc, word) {
		return function(expr1, expr2) {
			var that = exprbase(), prf;
			prf = function(expr, nassoc, before, after) {
				var r;
				if(priority > expr.priority() ||
						(priority === expr.priority() && assoc === nassoc)) {
					r = '(' + expr.action() + ')';
				} else {
					r = before + expr.action() + after;
				}
				return r;
			};
			that.priority = function() {
				return priority;
			};
			that.assoc = function() {
				return assoc;
			};
			that.action = function() {
				var r = '', spc = word ? ' ' : '';
				r += prf(expr1, 'right', '', spc);
				r += op;
				r += prf(expr2, 'left', spc, '');
				return r;
			};
			return that;
		};
	};
	J.prefixOperator = function(op, priority, word) {
		return function(expr1) {
			var that = exprbase();
			that.priority = function() {
				return priority;
			};
			that.action = function() {
				var r = '';
				r += op;
				r += unaryop(priority, expr1, word);
				return r;
			};
			return that;
		};
	};
	J.postfixOperator = function(op, priority) {
		return function(expr1) {
			var that = exprbase();
			that.priority = function() {
				return priority;
			};
			that.action = function() {
				var r = '';
				r += unaryop(priority, expr1);
				r += op;
				return r;
			};
			return that;
		};
	};
	J.ternaryOperator = function(priority) {
		return function(expr1, expr2, expr3) {
			var that = exprbase();
			that.priority = function() {
				return priority;
			}
			that.action = function() {
				var r = '';
				r += unaryop(priority, expr1);
				r += '?';
				r += expr2.action();
				r += ':';
				r += expr3.action();
				return r;
			};
			return that;
		};
	};
	J.newOperator = function(priority, list) {
		return function(expr, elist) {
			var that = exprbase();
			that.priority = function() {
				return priority;
			};
			that.action = function() {
				var r = '', i, d = '';
				r += 'new' + unaryop(priority, expr, true);
				if(list) {
					r += '(';
					for(i = 0; i < elist.length; i++) {
						r += d + elist[i].action();
						d = ',';
					}
					r += ')';
				}
				return r;
			};
			return that;
		};
	};
	J.callOperator = function(priority) {
		return function(expr1, elist) {
			var that = exprbase();
			that.isCall = function() {
				return expr1.isFunc();
			};
			that.priority = function() {
				return priority;
			};
			that.action = function(isIIFE) {
				var r = '', i, d = '';
				if(isIIFE) {
					r += '(' + unaryop(priority, expr1) + ')';
				} else {
					r += unaryop(priority, expr1);
				}
				r += '(';
				for(i = 0; i < elist.length; i++) {
					r += d + elist[i].action();
					d = ',';
				}
				r += ')';
				return r;
			};
			return that;
		};
	};
	J.indexOperator = function(priority) {
		return function(expr1, eindex) {
			var that = exprbase();
			that.priority = function() {
				return priority;
			};
			that.action = function() {
				var r = '', i, d = '';
				r += unaryop(priority, expr1);
				r += '[';
				r += eindex.action();
				r += ']';
				return r;
			};
			return that;
		};
	};
	J.expr = {};
	J.expr[',']    = J.infixOperator(',',   1000, 'left');
	J.expr['=']    = J.infixOperator('=',   1010, 'right');
	J.expr['+=']   = J.infixOperator('+=',  1010, 'right');
	J.expr['-=']   = J.infixOperator('-=',  1010, 'right');
	J.expr['*=']   = J.infixOperator('*=',  1010, 'right');
	J.expr['/=']   = J.infixOperator('/=',  1010, 'right');
	J.expr['%=']   = J.infixOperator('%=',  1010, 'right');
	J.expr['<<=']  = J.infixOperator('<<=', 1010, 'right');
	J.expr['>>=']  = J.infixOperator('>>=', 1010, 'right');
	J.expr['**=']  = J.infixOperator('**=', 1010, 'right');
	J.expr['>>>=']  = J.infixOperator('>>>=', 1010, 'right');
	J.expr['&=']   = J.infixOperator('&=',  1010, 'right');
	J.expr['^=']   = J.infixOperator('^=',  1010, 'right');
	J.expr['|=']   = J.infixOperator('|=',  1010, 'right');
	J.expr['...']  = J.prefixOperator('...', 1012, true);
	J.expr['yield']  = J.prefixOperator('yield', 1015, true);
	J.expr['yield*'] = J.prefixOperator('yield*', 1015, true);
	J.expr['?:']   = J.ternaryOperator(1018);
	J.expr['||']   = J.infixOperator('||',  1020, 'left');
	J.expr['&&']   = J.infixOperator('&&',  1030, 'left');
	J.expr['|']    = J.infixOperator('|',   1040, 'left');
	J.expr['^']    = J.infixOperator('^',   1050, 'left');
	J.expr['&']    = J.infixOperator('&',   1060, 'left');
	J.expr['==']   = J.infixOperator('==',  1070, 'left');
	J.expr['!=']   = J.infixOperator('!=',  1070, 'left');
	J.expr['===']  = J.infixOperator('===', 1070, 'left');
	J.expr['!==']  = J.infixOperator('!==', 1070, 'left');
	J.expr['<']    = J.infixOperator('<',   1080, 'left');
	J.expr['<=']   = J.infixOperator('<=',  1080, 'left');
	J.expr['>']    = J.infixOperator('>',   1080, 'left');
	J.expr['>=']   = J.infixOperator('>=',  1080, 'left');
	J.expr['instanceof'] = J.infixOperator('instanceof',  1080, 'left', true);
	J.expr['in']   = J.infixOperator('in',  1080, 'left', true);
	J.expr['>>']   = J.infixOperator('>>',  1090, 'left');
	J.expr['<<']   = J.infixOperator('<<',  1090, 'left');
	J.expr['>>>']  = J.infixOperator('>>>', 1090, 'left');
	J.expr['+']    = J.infixOperator('+',   1100, 'left');
	J.expr['-']    = J.infixOperator('-',   1100, 'left');
	J.expr['*']    = J.infixOperator('*',   1110, 'left');
	J.expr['/']    = J.infixOperator('/',   1110, 'left');
	J.expr['%']    = J.infixOperator('%',   1110, 'left');
	J.expr['**']   = J.infixOperator('**',  1115, 'right');
	J.expr['u-']   = J.prefixOperator('-',   1120);
	J.expr['u+']   = J.prefixOperator('+',   1120);
	J.expr['~']    = J.prefixOperator('~',   1120);
	J.expr['!']    = J.prefixOperator('!',   1120);
	J.expr['pr++'] = J.prefixOperator('++',  1120);
	J.expr['pr--'] = J.prefixOperator('--',  1120);
	J.expr['typeof'] = J.prefixOperator('typeof', 1120, true);
	J.expr['void']   = J.prefixOperator('void', 1120, true);
	J.expr['delete'] = J.prefixOperator('delete', 1120, true);
	J.expr['po++'] = J.postfixOperator('++',  1130);
	J.expr['po--'] = J.postfixOperator('--',  1130);
	J.expr['()']   = J.callOperator(1150);
	J.expr['new0'] = J.newOperator(1150, false);
	J.expr['[]']   = J.indexOperator(1150);
	J.expr['.']    = J.infixOperator('.',   1150, 'left');
	J.expr['new']  = J.newOperator(1150, true);
	J.numberValue = function(val) {
		var that = exprbase();
		that.priority = function() {
			return 9999999;
		};
		that.action = function() {
			return val + '';
		};
		return that;
	};
	J.variableValue = function(val) {
		var that = exprbase();
		that.priority = function() {
			return 9999999;
		};
		that.action = function() {
			return val;
		};
		return that;
	};
	J.stringValue = function(val, quotech) {
		var that = exprbase();
		that.priority = function() {
			return 9999999;
		};
		that.action = function() {
			return quotech + val + quotech;
		};
		return that;
	};
	J.arrayValue = function(elems) {
		var that = exprbase();
		that.priority = function() {
			return 9999999;
		};
		that.action = function() {
			var r = '[', i, d = '';
			for(i = 0; i < elems.length; i++) {
				r += d + elems[i].action();
				d = ',';
			}
			return r + ']';
		};
		return that;
	};
	J.objectValue = function(obj) {
		var that = exprbase();
		that.isParen = function() {
			return true;
		};
		that.priority = function() {
			return 9999999;
		};
		that.action = function() {
			var r = '{', i, d = '';
			for(i = 0; i < obj.length; i++) {
				r += d + obj[i][0].action();
				r += ':' + obj[i][1].action();
				d = ',';
			}
			return r + '}';
		};
		return that;
	};
	J.regexValue = function(val, flags) {
		var that = exprbase();
		that.priority = function() {
			return 9999999;
		};
		that.action = function() {
			var r = '/' + val + '/';
			if(flags) {
				r += flags;
			}
			return r;
		};
		return that;
	};
	J.functionValue = function(funcobj) {
		var that = exprbase();
		that.isFunc = function() {
			return true;
		};
		that.priority = function() {
			return 9999999;
		};
		that.action = function() {
			var r = 'function', i;
			if(funcobj.fname) {
				r += ' ';
				r += funcobj.fname;
			}
			r += '(';
			for(i = 0; i < funcobj.args.length; i++) {
				r += funcobj.args[i];
				if(i < funcobj.args.length - 1) {
					r += ',';
				}
			}
			r += '){';
			r += J.stmtlist(funcobj.slist);
			r += '}';
			return r;
		};
		return that;
	};
	stmtbase = function() {
		return {
			type: 'stmt',
			isIIFE: function() {
				return false;
			},
			isParen: function() {
				return false;
			},
			beforeDelim: function() {
				return ' ';
			},
			afterDelim: function() {
				return ';';
			}
		};
	};
	J.stmtif = function(cond, ifstmt, elsestmt) {
		var that = stmtbase();
		that.action = function() {
			var r = 'if(';
			r += cond.action() + ')';
			r += ifstmt.action();
			if(elsestmt) {
				r += ifstmt.afterDelim();
				r += 'else';
				r += elsestmt.beforeDelim();
				r += elsestmt.action();
			}
			return r;
		};
		return that;
	};
	J.stmtwhile = function(cond, stmt) {
		var that = stmtbase();
		that.action = function() {
			var r = 'while(';
			r += cond.action() + ')';
			r += stmt.action();
			return r;
		};
		return that;
	};
	J.stmtdo = function(cond, stmt) {
		var that = stmtbase();
		that.action = function() {
			var r = 'do{';
			r += stmt.action() + '}';
			r += 'while(' + cond.action() + ')';
			return r;
		};
		return that;
	};
	J.stmtfor = function(init, cond, next, stmt) {
		var that = stmtbase();
		that.action = function() {
			var r = 'for(';
			if(init) {
				r += init.action();
			}
			r += ';';
			if(cond) {
				r += cond.action();
			}
			r += ';';
			if(next) {
				r += next.action();
			}
			r += ')';
			r += stmt.action();
			return r;
		};
		return that;
	};
	J.stmtforin = function(name, expr, stmt) {
		var that = stmtbase();
		that.action = function() {
			var r = 'for(';
			r += name;
			r += ' in';
			r += J.actionexpr(expr);
			r += ')';
			r += stmt.action();
			return r;
		};
		return that;
	};
	J.stmtbreak = function(lbl) {
		var that = stmtbase();
		that.action = function() {
			var r = 'break';
			if(lbl) {
				r += ' ' + lbl;
			}
			return r;
		};
		return that;
	};
	J.stmtcontinue = function(lbl) {
		var that = stmtbase();
		that.action = function() {
			var r = 'continue';
			if(lbl) {
				r += ' ' + lbl;
			}
			return r;
		};
		return that;
	};
	J.stmtblock = function(slist) {
		var that = stmtbase();
		that.type = 'stmt';
		that.action = function() {
			return J.stmtlist(slist, '{', '}');
		};
		that.beforeDelim = function() {
			return slist.length === 1 ? slist[0].beforeDelim() : '';
		};
		that.afterDelim = function() {
			return slist.length === 1 ? slist[0].afterDelim() : '';
		};
		return that;
	};
	J.stmtreturn = function(expr) {
		var that = stmtbase();
		that.action = function() {
			var r = 'return';
			if(expr !== null) {
				r += J.actionexpr(expr);
			}
			return r;
		};
		return that;
	};
	J.stmtthrow = function(expr) {
		var that = stmtbase();
		that.action = function() {
			var r = 'throw';
			if(expr !== null) {
				r += J.actionexpr(expr);
			}
			return r;
		};
		return that;
	};
	J.stmtsimple = function(expr) {
		var that = stmtbase();
		that.isIIFE = function() {
			return expr.isCall();
		};
		that.isParen = function() {
			return expr.isParen();
		};
		that.action = function() {
			return expr.action();
		};
		that.actionIIFE = function() {
			return expr.action(true);
		};
		return that;
	};
	J.stmttry = function(body, vname, catchstmt, finstmt) {
		var that = stmtbase(), bra;
		bra = function(body) {
			if(body.beforeDelim() !== '') {
				return '{' + body.action() + '}';
			} else {
				return body.action();
			}
		};
		that.action = function() {
			var r = 'try';
			r += bra(body);
			if(catchstmt) {
				r += 'catch(' + vname + ')';
				r += bra(catchstmt);
			}
			if(finstmt) {
				r += 'finally';
				r += bra(finstmt);
			}
			return r;
		};
		that.afterDelim = function() {
			return '';
		};
		return that;
	};
	J.stmtswitch = function(expr, cllist) {
		var that = stmtbase();
		that.action = function() {
			var r = 'switch(' + expr.action() + '){', i, j;
			for(i = 0; i < cllist.length; i++) {
				if(cllist[i].expr) {
					r += 'case' + J.actionexpr(cllist[i].expr) + ':';
				} else {
					r += 'default:';
				}
				for(j = 0; j < cllist[i].stmt.length; j++) {
					r += cllist[i].stmt[j].action();
					r += cllist[i].stmt[j].afterDelim();
				}
			}
			r += '}';
			return r;
		};
		that.afterDelim = function() {
			return '';
		};
		return that;
	};
	J.stmtvar = function(vlist) {
		var that = stmtbase();
		that.action = function() {
			var r = 'var ', i;
			for(i = 0; i < vlist.length; i++) {
				r += vlist[i].name;
				if(vlist[i].expr !== null) {
					r += '=';
					r += vlist[i].expr.action();
				}
				if(i < vlist.length - 1) {
					r += ',';
				}
			}
			return r;
		}
		return that;
	};
	J.stmtlabel = function(name, stmt) {
		var that = stmtbase();
		that.action = function() {
			var r = '';
			r += name;
			r += ':';
			r += stmt.action();
			return r;
		};
		return that;
	};
	J.stmtfunction = function(funcobj) {
		var that = stmtbase();
		that.action = function() {
			var r = 'function', i;
			if(funcobj.fname) {
				r += ' ';
				r += funcobj.fname;
			}
			r += '(';
			for(i = 0; i < funcobj.args.length; i++) {
				r += funcobj.args[i];
				if(i < funcobj.args.length - 1) {
					r += ',';
				}
			}
			r += '){';
			r += J.stmtlist(funcobj.slist);
			r += '}';
			return r;
		};
		return that;
	};
	J.stmtlist = function(slist, ld, rd) {
		var r = '', i, expr;
		if(slist.length === 1) {
			if(slist[0].isIIFE()) {
				r += slist[0].actionIIFE();
			} else if(slist[0].isParen()) {
				r = '(' + slist[0].action() + ')';
			} else {
				r = slist[0].action();
			}
		} else {
			if(ld) {
				r += ld;
			}
			for(i = 0; i < slist.length; i++) {
				if(slist[i].isIIFE()) {
					r += slist[i].actionIIFE();
				} else if(slist[i].isParen()) {
					r += '(' + slist[i].action() + ')';
				} else {
					r += slist[i].action();
				}
				if(i < slist.length - 1) {
					r += slist[i].afterDelim();
				}
			}
			if(rd) {
				r += rd;
			}
		}
		return r;
	};
	J.actionexpr = function(expr) {
		var er = expr.action();
		if(J.exprFirst.test(er)) {
			return er;
		} else {
			return ' ' + er;
		}
	};
	J.exprFirst = /^[\(\)\+\-~!\.\'\"`/\[\]\{]/;
})(jsmin);
